@@ -17,14 +17,21 @@ import pandas as pd
 from bam_filter.sam_utils import process_bam, filter_reference_BAM
 import numpy as np
 from bam_filter.utils import get_arguments, create_output_files
+import json
 
 log = logging.getLogger("my_logger")
 
 
+def obj_dict(obj):
+    return obj.__dict__
+
+
 def main():
 
     logging.basicConfig(
-        level=logging.DEBUG, format="%(levelname)s ::: %(asctime)s ::: %(message)s"
+        level=logging.DEBUG,
+        format="%(levelname)s ::: %(asctime)s ::: %(message)s",
+        datefmt="%Y-%m-%d %H:%M:%S",
     )
 
     args = get_arguments()
@@ -35,10 +42,20 @@ def main():
     out_files = create_output_files(prefix=args.prefix, bam=args.bam)
 
     data = process_bam(
-        bam=args.bam, threads=args.threads, reference_lengths=args.reference_lengths
+        bam=args.bam,
+        threads=args.threads,
+        reference_lengths=args.reference_lengths,
+        scale=args.scale,
     )
 
+    data = list(filter(None, data))
+
     data_df = pd.DataFrame([x.to_summary() for x in data])
+    if args.read_length_freqs:
+        lens = [x.get_read_length_freqs() for x in data]
+        lens = json.dumps(lens, default=obj_dict, ensure_ascii=False, indent=4)
+    with open(out_files["read_length_freqs"], "w", encoding="utf-8") as outfile:
+        print(lens, file=outfile)
 
     logging.info(f"Writing reference statistics to {out_files['stats']}")
     data_df.to_csv(out_files["stats"], sep="\t", index=False, compression="gzip")
