@@ -14,7 +14,7 @@ import pyranges as pr
 
 log = logging.getLogger("my_logger")
 
-sys.setrecursionlimit(10 ** 6)
+sys.setrecursionlimit(10**6)
 
 # Function to calculate evenness of coverage
 def coverage_evenness(coverage):
@@ -76,7 +76,7 @@ def create_pyranges(reference, starts, ends, strands):
     )
 
 
-def get_bam_stats(params, ref_lengths=None):
+def get_bam_stats(params, ref_lengths=None, scale=1e6):
     """
     Worker function per chromosome
     loop over a bam file and create tuple with lists containing metrics:
@@ -156,7 +156,7 @@ def get_bam_stats(params, ref_lengths=None):
         mean_covered_bases = np.mean(ranges)
         bases_covered = int(len(cov_pos))
         # get SD from covered bases
-        cov_sd = np.std(cov_pos)
+        cov_sd = np.std(cov_pos, ddof=1)
         # get average coverage
         mean_coverage = sum(cov_pos) / reference_length
         mean_coverage_covered = sum(cov_pos) / bases_covered
@@ -170,6 +170,9 @@ def get_bam_stats(params, ref_lengths=None):
         c_v = cov_sd / mean_coverage
         read_mapq = [np.nan if x == 255 else x for x in read_mapq]
 
+        tax_abund_aln = round((n_alns / reference_length) * scale)
+        tax_abund_read = round((len(set(read_names)) / reference_length) * scale)
+
         log.debug(f"Number of reads: {len(set(read_names)):,}")
         log.debug(f"Number of alignments: {n_alns:,}")
         log.debug(f"Bases covered: {bases_covered:,}")
@@ -185,6 +188,8 @@ def get_bam_stats(params, ref_lengths=None):
         log.debug(f"C_v: {c_v:.2f}")
         log.debug(f"Mean mapq: {np.mean(read_mapq):.2f}")
         log.debug(f"GC content: {gc_content:.2f}")
+        log.debug(f"Taxonomic abundance (alns): {tax_abund_aln:.2f}")
+        log.debug(f"Taxonomic abundance (reads): {tax_abund_read:.2f}")
         data = BamAlignment(
             reference=reference,
             n_alns=n_alns,
@@ -210,34 +215,39 @@ def get_bam_stats(params, ref_lengths=None):
             mapping_quality=read_mapq,
             read_names=set(read_names),
             read_aln_score=read_aln_score,
+            tax_abund_aln=tax_abund_aln,
+            tax_abund_read=tax_abund_read,
         )
     else:
-        data = BamAlignment(
-            reference=reference,
-            n_alns=n_alns,
-            reference_length=reference_length,
-            bam_reference_length=bam_reference_length,
-            mean_coverage=np.nan,
-            mean_coverage_covered=np.nan,
-            bases_covered=np.nan,
-            max_covered_bases=np.nan,
-            mean_covered_bases=np.nan,
-            cov_evenness=np.nan,
-            breadth=np.nan,
-            exp_breadth=np.nan,
-            breadth_exp_ratio=np.nan,
-            c_v=np.nan,
-            edit_distances=np.nan,
-            # edit_distances_md=np.nan,
-            ani_nm=np.nan,
-            # ani_md=np.nan,
-            read_length=np.nan,
-            read_gc_content=np.nan,
-            read_aligned_length=np.nan,
-            mapping_quality=np.nan,
-            read_names=read_names,
-            read_aln_score=read_aln_score,
-        )
+        # data = BamAlignment(
+        #     reference=reference,
+        #     n_alns=n_alns,
+        #     reference_length=reference_length,
+        #     bam_reference_length=bam_reference_length,
+        #     mean_coverage=np.nan,
+        #     mean_coverage_covered=np.nan,
+        #     bases_covered=np.nan,
+        #     max_covered_bases=np.nan,
+        #     mean_covered_bases=np.nan,
+        #     cov_evenness=np.nan,
+        #     breadth=np.nan,
+        #     exp_breadth=np.nan,
+        #     breadth_exp_ratio=np.nan,
+        #     c_v=np.nan,
+        #     edit_distances=np.nan,
+        #     # edit_distances_md=np.nan,
+        #     ani_nm=np.nan,
+        #     # ani_md=np.nan,
+        #     read_length=np.nan,
+        #     read_gc_content=np.nan,
+        #     read_aligned_length=np.nan,
+        #     mapping_quality=np.nan,
+        #     read_names=read_names,
+        #     read_aln_score=read_aln_score,
+        #     tax_abund_aln=np.nan,
+        #     tax_abund_read=np.nan,
+        # )
+        data = None
     return data
 
 
@@ -272,6 +282,8 @@ class BamAlignment:
         cov_evenness,
         read_names,
         read_aln_score,
+        tax_abund_aln,
+        tax_abund_read,
     ):
         self.reference = reference
         self.n_alns = n_alns
@@ -297,6 +309,8 @@ class BamAlignment:
         self.c_v = c_v
         self.cov_evenness = cov_evenness
         self.read_names = read_names
+        self.tax_abund_aln = tax_abund_aln
+        self.tax_abund_read = tax_abund_read
         # function to convert class to dict
 
     def as_dict(self):
@@ -324,6 +338,8 @@ class BamAlignment:
             "breadth_exp_ratio": self.breadth_exp_ratio,
             "c_v": self.c_v,
             "cov_evenness": self.cov_evenness,
+            "tax_abund_read": self.tax_abund_read,
+            "tax_abund_aln": self.tax_abund_aln,
         }
 
     def to_summary(self):
@@ -331,19 +347,26 @@ class BamAlignment:
             warnings.simplefilter("ignore", category=RuntimeWarning)
             read_length_mean = np.mean(self.read_length)
             read_length_median = np.median(self.read_length)
+            read_length_std = np.std(self.read_length, ddof=1)
+            read_length_max = np.max(self.read_length)
+            read_length_min = np.min(self.read_length)
             read_length_mode = stats.mode(self.read_length)[0][0]
             read_aligned_length = np.mean(self.read_aligned_length)
             read_aln_score = np.mean(self.read_aln_score)
             mapping_quality = np.mean(self.mapping_quality)
             edit_distances = np.mean(self.edit_distances)
             read_ani_mean = np.mean(self.ani_nm)
+            read_ani_std = np.std(self.ani_nm, ddof=1)
+            reads_ani_median = np.median(self.ani_nm)
             gc_content = (np.sum(self.read_gc_content) / np.sum(self.read_length)) * 100
-
         return {
             "reference": self.reference,
             "n_reads": len(self.read_names),
             "n_alns": self.n_alns,
             "read_length_mean": read_length_mean,
+            "read_length_std": read_length_std,
+            "read_length_min": read_length_min,
+            "read_length_max": read_length_max,
             "read_length_median": read_length_median,
             "read_length_mode": read_length_mode,
             "gc_content": gc_content,
@@ -353,6 +376,8 @@ class BamAlignment:
             "edit_distances": edit_distances,
             # "edit_distances_md": np.mean(self.edit_distances_md),
             "read_ani_mean": read_ani_mean,
+            "read_ani_std": read_ani_std,
+            "reads_ani_median": reads_ani_median,
             # "ani_md": np.mean(self.ani_md),
             "bases_covered": self.bases_covered,
             "max_covered_bases": self.max_covered_bases,
@@ -366,11 +391,21 @@ class BamAlignment:
             "breadth_exp_ratio": self.breadth_exp_ratio,
             "c_v": self.c_v,
             "cov_evenness": self.cov_evenness,
+            "tax_abund_read": self.tax_abund_read,
+            "tax_abund_aln": self.tax_abund_aln,
         }
 
+    def get_read_length_freqs(self):
+        frags = {}
+        lengths = pd.Series(self.read_length)
+        lengths = lengths.value_counts().sort_index()
+        freqs = list(lengths / sum(lengths))
+        frags[self.reference] = {"length": list(lengths.index), "freq": freqs}
+        return frags
+
 
 # Inspired from https://gigabaseorgigabyte.wordpress.com/2017/04/14/getting-the-edit-distance-from-a-bam-alignment-a-journey/
-def process_bam(bam, threads=1, reference_lengths=None):
+def process_bam(bam, threads=1, reference_lengths=None, scale=1e6):
     """
     Processing function: calls pool of worker functions
     to extract from a bam file two definitions of the edit distances to the reference genome scaled by read length
@@ -379,16 +414,35 @@ def process_bam(bam, threads=1, reference_lengths=None):
     logging.info(f"Loading BAM file")
     save = pysam.set_verbosity(0)
     samfile = pysam.AlignmentFile(bam, "rb")
+
+    references = samfile.references
+
+    chr_lengths = []
+    for chrom in samfile.references:
+        chr_lengths.append(samfile.get_reference_length(chrom))
+    max_chr_length = np.max(chr_lengths)
     pysam.set_verbosity(save)
+    ref_lengths = None
 
     if reference_lengths is not None:
         ref_lengths = pd.read_csv(
             reference_lengths, sep="\t", index_col=0, names=["reference", "length"]
         )
+        # check if the dataframe contains all the References in the BAM file
+        if not set(references).issubset(set(ref_lengths.index)):
+            logging.error(
+                f"The BAM file contains references not found in the reference lengths file"
+            )
+            sys.exit(1)
+        max_chr_length = np.max(ref_lengths["length"].tolist())
 
     if not samfile.has_index():
         logging.info(f"BAM index not found. Indexing...")
-        pysam.index(bam)
+        if max_chr_length > 536870912:
+            logging.info(f"A reference is longer than 2^29, indexing with csi")
+            pysam.index(bam, "-c")
+        else:
+            pysam.index(bam)
 
         logging.info(f"Reloading BAM file")
         samfile = pysam.AlignmentFile(
@@ -397,21 +451,27 @@ def process_bam(bam, threads=1, reference_lengths=None):
 
     total_refs = samfile.nreferences
     logging.info(f"Found {total_refs:,} reference sequences")
-    logging.info(f"Found {samfile.mapped:,} alignments")
+    # logging.info(f"Found {samfile.mapped:,} alignments")
 
-    references = samfile.references[0:100]
     params = zip([bam] * len(references), references)
     try:
         logging.info(f"Getting stats for each reference...")
 
         if is_debug():
             data = list(
-                map(functools.partial(get_bam_stats, ref_lengths=ref_lengths), params)
+                map(
+                    functools.partial(
+                        get_bam_stats, ref_lengths=ref_lengths, scale=scale
+                    ),
+                    params,
+                )
             )
         else:
 
             p = Pool(
-                threads, initializer=initializer, initargs=([params, ref_lengths],)
+                threads,
+                initializer=initializer,
+                initargs=([params, ref_lengths, scale],),
             )
             c_size = calc_chunksize(
                 n_workers=threads,
@@ -420,7 +480,9 @@ def process_bam(bam, threads=1, reference_lengths=None):
             data = list(
                 tqdm.tqdm(
                     p.imap_unordered(
-                        functools.partial(get_bam_stats, ref_lengths=ref_lengths),
+                        functools.partial(
+                            get_bam_stats, ref_lengths=ref_lengths, scale=scale
+                        ),
                         params,
                         chunksize=c_size,
                     ),
@@ -463,68 +525,87 @@ def filter_reference_BAM(bam, df, filter_conditions, threads, out_files, sort_me
         & (df["breadth_exp_ratio"] >= filter_conditions["min_expected_breadth_ratio"])
         & (df["cov_evenness"] >= filter_conditions["min_coverage_evenness"])
     ]
+    if len(df_filtered.index) > 0:
+        refs_dict = dict(zip(df_filtered["reference"], df_filtered["reference_length"]))
+        (ref_names, ref_lengths) = zip(*refs_dict.items())
 
-    refs_dict = dict(zip(df_filtered["reference"], df_filtered["reference_length"]))
-    (ref_names, ref_lengths) = zip(*refs_dict.items())
+        out_bam_file = pysam.Samfile(
+            out_files["bam_filtered_tmp"],
+            "wb",
+            referencenames=list(ref_names),
+            referencelengths=list(ref_lengths),
+            threads=threads,
+        )
+        header = pysam.AlignmentHeader.from_references(
+            list(ref_names), list(ref_lengths)
+        )
+        logging.info(f"Writing filtered BAM file... (be patient)")
+        references = df_filtered["reference"].values
+        params = zip([bam] * len(references), references)
+        logging.info(f"Saving filtered stats...")
+        df_filtered.to_csv(
+            out_files["stats_filtered"], sep="\t", index=False, compression="gzip"
+        )
+        try:
+            logging.info(f"Filtering BAM file...")
 
-    out_bam_file = pysam.Samfile(
-        out_files["bam_filtered_tmp"],
-        "wb",
-        referencenames=list(ref_names),
-        referencelengths=list(ref_lengths),
-        threads=threads,
-    )
-    header = pysam.AlignmentHeader.from_references(list(ref_names), list(ref_lengths))
-    logging.info(f"Writing filtered BAM file... (be patient)")
-    references = df_filtered["reference"].values
-    params = zip([bam] * len(references), references)
-    logging.info(f"Saving filtered stats...")
-    df_filtered.to_csv(
-        out_files["stats_filtered"], sep="\t", index=False, compression="gzip"
-    )
-    try:
-        logging.info(f"Filtering BAM file...")
+            if is_debug():
+                alns = list(map(get_alns, params))
+            else:
 
-        if is_debug():
-            alns = list(map(get_alns, params))
-        else:
-
-            p = Pool(threads)
-            c_size = calc_chunksize(threads, len(references))
-            alns = list(
-                tqdm.tqdm(
-                    p.imap_unordered(get_alns, params, chunksize=c_size),
-                    total=len(references),
-                    leave=False,
-                    ncols=80,
-                    desc=f"References processed",
+                p = Pool(threads)
+                c_size = calc_chunksize(threads, len(references))
+                alns = list(
+                    tqdm.tqdm(
+                        p.imap_unordered(get_alns, params, chunksize=c_size),
+                        total=len(references),
+                        leave=False,
+                        ncols=80,
+                        desc=f"References processed",
+                    )
                 )
-            )
 
-            p.close()
+                p.close()
+                p.join()
+
+        except KeyboardInterrupt:
+            logging.info(f"User canceled the operation. Terminating jobs.")
+            p.terminate()
             p.join()
+            sys.exit()
 
-    except KeyboardInterrupt:
-        logging.info(f"User canceled the operation. Terminating jobs.")
-        p.terminate()
-        p.join()
-        sys.exit()
+        for aln in fast_flatten(alns):
+            out_bam_file.write(pysam.AlignedSegment.fromstring(aln, header=header))
+        out_bam_file.close()
+        pysam.sort(
+            "-@",
+            str(threads),
+            "-m",
+            str(sort_memory),
+            "-o",
+            out_files["bam_filtered"],
+            out_files["bam_filtered_tmp"],
+        )
 
-    samfile = pysam.AlignmentFile(bam, "rb")
-    for aln in fast_flatten(alns):
-        out_bam_file.write(pysam.AlignedSegment.fromstring(aln, header=header))
-    out_bam_file.close()
-    pysam.sort(
-        "-@",
-        str(threads),
-        "-m",
-        str(sort_memory),
-        "-o",
-        out_files["bam_filtered"],
-        out_files["bam_filtered_tmp"],
-    )
-    pysam.index(out_files["bam_filtered"])
-    os.remove(out_files["bam_filtered_tmp"])
+        save = pysam.set_verbosity(0)
+        samfile = pysam.AlignmentFile(out_files["bam_filtered"], "rb")
+        chr_lengths = []
+        for chrom in samfile.references:
+            chr_lengths.append(samfile.get_reference_length(chrom))
+        max_chr_length = np.max(chr_lengths)
+        pysam.set_verbosity(save)
+        samfile.close()
+
+        logging.info(f"BAM index not found. Indexing...")
+        if max_chr_length > 536870912:
+            logging.info(f"A reference is longer than 2^29, indexing with csi")
+            pysam.index(out_files["bam_filtered"], "-c")
+        else:
+            pysam.index(out_files["bam_filtered"])
+
+        os.remove(out_files["bam_filtered_tmp"])
+    else:
+        logging.info(f"No references meet the filter conditions. Skipping...")
 
 
 def get_alns(params):
