@@ -50,15 +50,27 @@ def is_integer(n):
 
 # function to check if the input value has K, M or G suffix in it
 def check_suffix(val, parser, var):
-    units = ["K", "M", "G"]
+    if var == "--scale":
+        units = ["K", "M"]
+    else:
+        units = ["K", "M", "G"]
     unit = val[-1]
     value = int(val[:-1])
 
     if is_integer(value) & (unit in units) & (value > 0):
-        return val
+        if var == "--scale":
+            if unit == "K":
+                val = value * 1000
+            elif unit == "M":
+                val = value * 1000000
+            elif unit == "G":
+                val = value * 1000000000
+            return val
+        else:
+            return val
     else:
         parser.error(
-            "argument %s: Invalid value %s. Memory has to be an integer larger than 0 with the following suffix K, M or G"
+            "argument %s: Invalid value %s. Has to be an integer larger than 0 with the following suffix K, M or G"
             % (var, val)
         )
 
@@ -112,6 +124,7 @@ defaults = {
     "prefix": None,
     "sort_memory": "1G",
     "reference_lengths": None,
+    "scale": 1e6,
 }
 
 help_msg = {
@@ -124,6 +137,8 @@ help_msg = {
     "min_read_ani": "Minimum average read ANI",
     "min_coverage_evenness": "Minimum coverage evenness",
     "sort_memory": "Set maximum memory per thread for sorting; suffix K/M/G recognized",
+    "scale": "Scale taxonomic abundance by this factor; suffix K/M recognized",
+    "read_length_freqs": "Save a JSON file with the read length frequencies mapped to each reference",
     "help": "Help message",
     "debug": f"Print debug messages",
     "reference_lengths": "File with references lengths",
@@ -133,7 +148,7 @@ help_msg = {
 
 def get_arguments(argv=None):
     parser = argparse.ArgumentParser(
-        description="A simple tool to calculate metrics from a BAM file and filter references to be used with Woltka",
+        description="A simple tool to calculate metrics from a BAM file and filter with uneven coverage.",
         formatter_class=argparse.ArgumentDefaultsHelpFormatter,
     )
     parser.add_argument(
@@ -199,7 +214,7 @@ def get_arguments(argv=None):
         "-a",
         "--min-read-ani",
         type=lambda x: float(
-            check_values(x, minval=0, maxval=1, parser=parser, var="--min-read-ani")
+            check_values(x, minval=0, maxval=100, parser=parser, var="--min-read-ani")
         ),
         default=defaults["min_read_ani"],
         dest="min_read_ani",
@@ -225,6 +240,13 @@ def get_arguments(argv=None):
         dest="sort_memory",
         help=help_msg["sort_memory"],
     )
+    parser.add_argument(
+        "--scale",
+        type=lambda x: check_suffix(x, parser=parser, var="--scale"),
+        default=defaults["scale"],
+        dest="scale",
+        help=help_msg["scale"],
+    )
     # reference_lengths
     parser.add_argument(
         "-r",
@@ -234,6 +256,12 @@ def get_arguments(argv=None):
         dest="reference_lengths",
         help=help_msg["reference_lengths"],
     )
+    parser.add_argument(
+        "--read-length-freqs",
+        dest="read_length_freqs",
+        action="store_true",
+        help=help_msg["read_length_freqs"],
+    )
     parser.add_argument(
         "--debug", dest="debug", action="store_true", help=help_msg["debug"]
     )
@@ -375,5 +403,6 @@ def create_output_files(prefix, bam):
         "stats_filtered": f"{prefix}_stats-filtered.tsv.gz",
         "bam_filtered_tmp": f"{prefix}.filtered.tmp.bam",
         "bam_filtered": f"{prefix}.filtered.bam",
+        "read_length_freqs": f"{prefix}_read-length-freqs.json",
     }
     return out_files
